import { Box, Color, Static, Text, render } from 'ink'
import { transform, debounce } from 'lodash'
import { useEffect, useState } from 'react'
import chokidar from 'chokidar'
import glob from 'glob'
import h from 'react-hyperscript'
import merge from 'webpack-merge'
import rimraf from 'rimraf'
import webpack from 'webpack'

import { promises as fs } from 'fs'
import { promisify } from 'util'
import config from 'lib/config'
import prepack from './prepack'

import build from './build'
import copyHTML from './copyHTML'
import tree from './tree'

require('dotenv').config()

const pathsToAdd = [],
  isProduction = process.env.NODE_ENV == 'production',
  babelConfig = Object.assign(require('../babel.config.js'), { ignore: [] })

async function pack({ updateState }) {
  const entryFiles = glob.sync(`${config.prePackDir}/**/*.js`),
    entry = transform(
      entryFiles,
      (result, entry) => {
        result[entry.replace(config.prePackDir, '').replace('.js', '')] = entry
      },
      {},
    ),
    stats = await promisify(webpack)({
      mode: isProduction ? 'none' : 'development',
      entry,
      output: {
        path: config.destDir,
        filename: '[name]-[contenthash].js',
      },
      cache: {
        type: 'filesystem',
      },
      resolve: {
        alias: babelConfig.plugins.find(
          ([name]) => name == 'babel-plugin-module-resolver',
        )[1].alias,
      },
      module: {
        rules: [
          {
            test: /\.js$/,
            use: [{ loader: 'babel-loader', options: babelConfig }],
          },
        ],
      },
      plugins: [
        new webpack.EnvironmentPlugin({
          DEBUG: null,
          DEV_SERVER: null,
          NODE_ENV: process.env.NODE_ENV,
        }),
        new webpack.DefinePlugin({
          'process.env.CONFIG': JSON.stringify(config),
          'process.env.WEB': 'true',
          'process.env.SSR': 'false',
        }),
      ],
      optimization: isProduction
        ? {
          minimize: true,
        }
        : {},
    })

  updateState({ webpackStats: stats })
}

async function runBuild({ updateState }) {
  try {
    updateState({ status: 'clean...' })
    rimraf.sync(config.destDir)
    updateState({ status: 'build...' })
    await build(tree)
    updateState({ status: 'prepack...' })
    await prepack(tree)
    updateState({ status: 'pack...' })
    await fs.mkdir(config.destDir, { recursive: true })
    await pack({ updateState })
    updateState({ status: 'copyHTML...' })
    await copyHTML(tree)
  } catch (e) {
    updateState({ error: e })
    throw e
  }
}

function isEntryPath(path) {
  return path.endsWith('md')
}

async function initialize({ updateState, addPaths, expressApp }) {
  const watcher = chokidar.watch(`${config.srcDir}/**/*.md`)

  watcher.on('add', path => {
    if (isEntryPath(path)) pathsToAdd.push(path)
    addPaths()
  })

  if (process.env.DEV_SERVER) {
    const express = require('express'),
      http = require('http'),
      ws = require('ws'),
      app = expressApp || express(),
      server = http.createServer(app),
      wss = new ws.Server({ server, path: '/socket', clientTracking: true }),
      port = app.get('port') || process.env.PORT

    wss.on('connection', socket => socket.send('hi'))

    app.use(express.static(config.destDir))

    if (!port) throw new Error('Port was not configured')
    server.listen(port)
    updateState({serverStatus: `listening on ${port}`})

    watcher.on('ready', async () => {
      watcher.on('change', async path => {
        delete require.cache[`${process.cwd()}/${path}`]
        const start = Date.now()
        updateState({ status: 'rebuild...' })
        if (isEntryPath(path)) await tree.get(path).change()
        await runBuild({ updateState })
        updateState({
          status: `rebuild success, watch ${Date.now() - start}ms`,
          error: null,
        })

        setTimeout(() => {
          wss.clients.forEach(client => client.send('reload'))
        })
      })
    })
  }
}

function App({ expressApp }) {
  const [state, setState] = useState({
      status: 'starting',
      serverStatus: 'n/a',
      error: null,
      logMessages: ['started'],
      webpackStats: null,
    }),
    updateState = newValues => setState(state => merge(state, newValues)),
    addPaths = debounce(async function() {
      const promises = pathsToAdd.splice(0).map(tree.add),
        start = Date.now()

      await Promise.all(promises)
      await runBuild({ updateState })
      if (process.env.DEV_SERVER) {
        updateState({
          status: `first build success, watch ${Date.now() - start}ms`,
        })
      } else {
        updateState({ status: `built ${Date.now() - start}ms` })
        process.exit()
      }
    }, 10)

  useEffect(() => {
    process.on('unhandledRejection', error => {
      console.log('uncaught')
      updateState({ error })
    })

    process.on('uncaughtException', error => {
      console.log('uncaught')
      updateState({ error })
    })

    console.log = (...args) => {
      updateState({
        logMessages: [args.join(' ')],
      })
    }

    initialize({ updateState, addPaths, expressApp })
  }, [])

  return h(Box, { flexDirection: 'column' }, [
    h(
      Static,
      state.logMessages.map((message, i) =>
        h(Box, { key: `message-${i}` }, message),
      ),
    ),
    h(Box, `DevServer Status: ${state.serverStatus}`),
    h(Box, `Build Status: ${state.status}`),
    state.webpackStats
      ? h(
        Box,
        `\nWebpack stats: \n${state.webpackStats.toString({
          // context: '/',
          all: false,
          builtAt: true,
          entrypoints: true,
          timings: true,
          colors: true,
        })}`,
      )
      : null,
    state.error &&
      h(
        Color,
        { red: true },
        h(Text, { bold: true }, `Error: ${state.error.stack}`),
      ),
  ])
}

export function start({ expressApp }) {
  return render(h(App, { expressApp }), process.stdout)
}
